use IO::Handle;

#    STDERR->autoflush(1);
#    STDOUT->autoflush(1);

package Chainmail::MessageType::Rfc822;

require Chainmail::MessageType::Rfc822::FieldParser;
require Chainmail::Address;
use Chainmail::Report;
use Chainmail::Lib;
use Chainmail::Config;
use Socket;


@sender_fields = ( # From the sendmail bat book. In descending order of preference.
		  "Resent-Sender", 
		  "Resent-From", 
		  "Resent-Reply-To", 
		  "Sender",
		  "From", 
		  "Apparently-From",
		  "Reply-To", 
		  "Full-Name",
		  "Return-Receipt-To",
		 );

@reply_to_fields = ( "Reply-To", # picked +/- at random
		     "From", 
		     "Apparently-From", 
		     "Resent-From", 
		     "Sender", 
		     "Resent-Sender", 
		     "Return-Receipt-To" 
		   );

@error_recipient_fields = (
			   "Errors-To",
			   "Return-Path",
			   @sender_fields
		   );


@resent_recip_fields =		     
  ("Resent-To",
   "Resent-Cc",
   "Resent-Bcc"
  );

@recip_fields
=
  ( "To", 
    "Cc", 
    "Bcc",
    "Apparently-To"
  );

@recipient_fields = [
		     # Also from sendmail. 

		     # List of lists; if anything in a list matches
		     # then the following lists are not tried at all

		     @resent_recip_fields,
		     @recip_fields
		    ];


@nonrecipient_fields = ( "X-loop" );

@hide_fields = 
			     ( "Bcc",
			       "Resent-Bcc" );

@add_if_absent_fields = ( 
			 "Message-ID",
			 "Date",
#			 "X-Mailer"
			);

@replace_fields = ( # Only replace if present
		    # "Comment"
		   );

@prepend_fields = ( "Received" ); # Unconditional

@address_fields = 
  ( # Should be all address fields
   @sender_fields, 
   @error_recipient_fields,
   @resent_recip_fields,
   @recip_fields,
   @nonrecipient_fields
 );




sub new 
   {
        my $proto = shift;
        my $class = ref($proto) || $proto;

        my $self  = {};

        $self->{BODY}   = undef;
        $self->{HEADER} = undef;
        @{$self->{FIELD}}  = ();
    
        bless ($self, $class);
        return $self;
    }

sub parse
{
  my $self = shift;
  my $text = shift;

  my $header;

  if ( $text =~ /^from ([^\n]*)(?:\n)?(.*)/si )
    {
      inform( "UNIX mailbox style from line: \"$1\" is being ignored" );
      $text = $2;
    }

  ($self->{HEADER},$self->{BODY}) = split /\n\n/, $text, 2;
  
#  $self->{BODY} .= "\n"
#    unless $self->{BODY} =~ /\n$/s;

  $self->parse_header();

  return $self;
}

sub parse_header
  {
    my $self = shift;
    local $_;

    enter( "Parsing RFC822 header" );

    @{$self->{FIELD}} = ();

    $_ = $self->{HEADER};

    do
      {
	/^((?:[^\n]*(?:\n\s)?)*)\n?(.*)/s
	  || fatal_error( "regexp failed to find top field of header:\n\t$_" );
	
	$field = $1;
	$_ = $2;

	$field =~ /^([^:]+?)\s*:(.*)/s
	  || fatal_error( "regexp failed to split header field:\n\t$field" ) ;

	$name = $1;
	$value = $2;

	debug( "header-field $name is \"$value\"" );

	if ( member( $name,@replace_fields ) )
	  {
	    $value = $self->generate_field( $name );
	  }

	my $parser = undef;

	if ( member( $name,@address_fields ) )
	  {
	    $parser = Chainmail::MessageType::Rfc822::FieldParser->new();
	    $parser->transform_addresses( $value );
	  }

	push  @{$self->{FIELD}}, $self->_make_field( $name, $value, $parser );

      } while (length);

    my %add_fields;

    @add_fields{ @add_if_absent_fields } = @add_if_absent_fields;

#    foreach $name (@add_if_absent_fields) 
#      {
#	$add_fields{ $name } = '1';
#      }

    foreach my $field (@{$self->{FIELD}})
      {
	if ( member( $field->{NAME}, @replace_fields ) )
	  {
	    $field->{VALUE} = $self->generate_field( $field->{NAME} );
	  }
	if ( $name = member( $field->{NAME}, keys %add_fields ) )
	  {

	    debug( "field $name already present" );
	    debug( "field $name does not exist") if !exists($add_fields{ $name });
	    
	    delete $add_fields{ $name }; # keep the case
	  }
      }

    foreach my $name (keys %add_fields)
      {
	push @{$self->{FIELD}}, $self->_make_field( $name, $self->generate_field( $name ) );
      }

    foreach my $name (@prepend_fields)
      {
	unshift @{$self->{FIELD}}, 
		$self->_make_field( $name, $self->generate_field( $name ) ) ;
      }

    leave();
  }


sub get_contents
  {
    my $self = shift;

    return $self->construct_header() . "\n" . $self->{BODY};
  }
    
sub construct_header
  {
    my $self = shift;
    my $text = "";

    $self->parse_header()
      if !defined @{$self->{FIELD}};

    foreach my $field ( @{$self->{FIELD}} )
      {
	error( "empty field address $field found when re-constructing header") 
	  if (!defined $field || !defined $field->{NAME});
	
	$text .= $field->{NAME} . ": ";
	
	if( defined $field->{PARSER} )
	  {
	    $text .=  $field->{PARSER}->get_pretty_text();
	  }
	else
	  {
	    my $line .= $field->{VALUE};
	    $line =~ s/^[\s]*//;
	    $text .= $line;
	  }
	
	$text .= "\n";
      }

    return $text;
  }

sub get_senders
  {
    my $self = shift;

    @_ = $self->get_multifield_addresses( @sender_fields );

    return @_;
  }

sub get_reply_to
  {
    my $self = shift;
    my @addrs = $self->get_multifield_addresses( @reply_to_fields );

    if (!(@addrs))
      {

	my ($name,$passwd,$uid,$gid, $quota,$comment,$gcos,$dir,$shell,$expire) 
	  = getpwuid($>);

	push @addrs, Chainmail::Address->new( $name, undef, $gecos );
      }

    return @addrs;
  }


sub get_recipients
  {
    my $self = shift;
    my $addr_list = ();
    my @addrs = ();
    

    foreach $addr_list  (@recipient_fields)
      {
	push @addrs, $self->get_multifield_addresses( @{$addr_list} );
      }

    if (! (@addrs) )
      {
	debug( "no recipient addresses found" );
	return undef;
      }

    my @nonaddrs = $self->get_multifield_addresses( @nonrecipient_fields );
    my @good_addrs;

    foreach my $recip (@addrs) 
      {
	foreach my $noperson (@nonaddrs)
	  {
	    goto dont_add if $recip->get_address() eq $noperson->get_address();
	  }
	push @good_addrs, $recip;
	next;
      dont_add:
	debug( $recip->get_address() . " was removed from recipient list" );
      }

    return @good_addrs;
  }


sub get_addresses
  {
    my $self = shift;
 
    my $field_name = lc( shift );

    my @addrs = ();

    foreach my $field (@{$self->{FIELD}})
      {

	if ( lc( $field->{NAME} ) eq $field_name ) 
	  {
	    debug( "getting addresses from field: $field_name" );

	    if ( (!defined ($field->{PARSER})) )
	      {
		error( "field $field_name apparently has not been parsed yet" );
		return;
	      }

	    @_ = $field->{PARSER}->grab_addresses();

	    foreach my $address ( @_  )
	      {
		debug( "adding address " . $address->get_text() . " to address list" );
		# This line is currently needed (1999 08 23), 
		# because otherwise perl (5.005) garbage collects the address.
		# Now, that's what I call a BUG (or have I misunderstood something?).
	      }
	    
	    push @addrs, @_
	      if( defined @_ && defined $_[0] );
	  }
      }

    return @addrs;
  }

sub get_multifield_addresses
  {
    my $self = shift;
    my @fields = @_;


    my @addrs;
    my $f;
    my @temp;

    foreach $f (@fields)
      {
	@temp = $self->get_addresses( $f );

	(push @addrs, @temp)
	  if ( @temp && defined( $temp[ 0 ] ));
      }
    
    return @addrs;
  }
  
sub generate_date
  {
    my @days_of_week = ( "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" );
    my @months = ( "Jan"  ,  "Feb" ,  "Mar"  ,  "Apr"
		   ,  "May"  ,  "Jun" ,  "Jul"  ,  "Aug"
		   ,  "Sep"  ,  "Oct" ,  "Nov"  ,  "Dec" );
    
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday) =
      gmtime();

    my $day = $days_of_week[ $wday ];
    my $date = $mday . " " . $months[ $mon ] . " " . 
      ($year+1900); 
    # Strictly by RFC822 year should be 2 digit, but nobody seems to care

    my $time = sprintf( "%02d", $hour ) . ":" . sprintf( "%02d", $min ) . ":" . sprintf( "%02d", $sec ) . " GMT";
    # TODO: get real timezone

    return "$day, $date $time";
  }

sub generate_field
  {
    my $self = shift;
    my $name = lc( shift );

    if ( $name eq "received" )
      {
	my $date = generate_date();
	my $domain = get_hostname();

	my $string = "by " . $domain . " (chainmail); $date";
	return $string;
      }

    if ( $name eq "date" )
      {
	return generate_date();
      }
    if ( $name eq "message-id" )
      {

	my $noconflict = crypt( rand(4294967296) , 
			    join '',('.', '/', 0..9, 'A'..'Z', 'a'..'z')
			    [rand 64, rand 64] ) . time;

	$noconflict =~ tr/a-zA-Z0-9//cd;
	return "<" . $noconflict . "-chainmail@" . get_hostname() .">";
      }

    return "chainmail - generate field is not yet implemented for $name";
  }

sub _make_field
  {
    my $self = shift;
    
    my $name= shift;
    my $value= shift;
    my $parser= shift;

    my $field;

    $field->{ NAME } = $name;
    $field->{ VALUE } = $value;
    $field->{ PARSER } = $parser
      if defined $parser;

    return $field;
  }
    
1;
    
